/*---------------------------------------------------------------------------*\
 =========                   |
 \\      /   F ield          | OpenFOAM: The Open Source CFD Toolbox
  \\    /    O peration      |
   \\  /     A nd            | Copyright (C) 1991-2005 OpenCFD Ltd.
    \\/      M anipulation   |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

Application
    makeAxialMesh

Description
    Creates a axial-symmetric grid from a two-dimensional grid

    Needs two inputs:
    - name of the boundary that is the symmetry axis
    - name of the boundary that is to be splitted into two wedge-boundaries (front and back)

    Assumes the following about the grid:
    - 2D-Grid (just one cell thick)
    - parallel to the XY-plane
    - the boundary that is the symmetry axis is a straight line
    - the "front" and "back" are one patch

    Afterwards: use the collapseEdges utility to remove the faces at the symmetry axis
 
\*---------------------------------------------------------------------------*/


#include "argList.H"
#include "Time.H"
#include "repatchPolyTopoChanger.H"

#include "plane.H"
#include "line.H"
#include "faceSet.H"
#include "SortableList.H"

#include "mathematicalConstants.H"

using namespace Foam;

// #define VERBOSE_OUTPUT

typedef line<point,point> linie;

// get the plane of the grid
// right now it assumes parallel to XY

plane getPlane(pointField& points) {
  scalar x=0,y=0,z=0;
  label  n=0;

  forAll(points,pointI) {
    point &pt=points[pointI];
    x += pt.x();    y += pt.y();    z += pt.z();
    n ++;
  }

  plane result(point(x/n,y/n,z/n),vector(0,0,1));

  return result;
}

// calculate the symmetry axis from the "symmetry patch"
// by finding the two points that are farthest apart

linie getAxis(const polyPatch &axis,plane &pl) {
  scalar dist=0;
  int cnt=0;
  point pt1=vector(0,0,0); //DPS explicity initializing these avoids compiler warnings about uninit. var.
  point pt2=vector(0,0,0); //DPS and no harm is done since case 0 and 1 overwrite them immediately
  const pointField &pts=axis.localPoints();

  forAll(pts,pointI) {
    switch(cnt) {
    case 0:
      pt1=pl.nearestPoint(pts[pointI]);
      break;
    case 1:
      pt2=pl.nearestPoint(pts[pointI]);
      dist=mag(pt1-pt2);
      break;
    default:
      point pt=pl.nearestPoint(pts[pointI]);
      if( mag(pt-pt1)>dist || mag(pt-pt2)>dist) {
	if( mag(pt-pt1) > mag(pt-pt2)) {
	  pt2=pt;
	} else {
	  pt1=pt;
	}
	dist=mag(pt1-pt2);
      }
    }
    cnt++;
  }

  return linie(pt1,pt2);
}

// angle (as suggested in the OpenFOAM-Userguide)
const scalar angle=2.5;   //DPS the plane is rotated +2.5 degrees and -2.5 degrees to make a 5 deg. wedge

// changes the coordinates of the points

/*
    This docu provided by E David Huckaby
    (plus a fix to a bug)

    Inputs:
        cutPlane - INFINITE plane (getPlane)
            - defined by position vector and normal vector
        axisLine - FINITE line (getAxis)
            - defined by two points
            - axis of rotation
        mesh - 
    

    Variables:
        "basePoint" is the projection of the cartesian point, "oldPoint"
            onto the cutting plane "cutPlane". This is equivalent to
            the nearest point from "oldPoint" to "cutPlane".
        "axisPoint" is the projection of "basePoint" along the axis, "axisLine".
            If "axisLine" was an INFINITE line then this would be equilivalent to the
            the neareast point from "axisLine" to "basePoint".  This is not the case,
            so the projection is done with a dot product ("&") with the unit vector in the
            direction of "axisLine", "axisDir".    
        "radius" is the radial coordinate of the cartesian point, in the plane
            perpendicular to the rotation axis.

*/

// #define VERBOSE_OUTPUT

void changeCoordinates(polyMesh &mesh,plane cutPlane,linie &axisLine) {
  repatchPolyTopoChanger topo(mesh);

  pointField oldPoints=mesh.points();                //DPS the old points will be rotated too
  pointField newPoints(oldPoints.size());

  const scalar factor=std::sin(0.5*angle/180.*mathematicalConstant::pi);
  vector axisDir = axisLine.vec()/axisLine.mag();      

  scalar minRadius=1e10,maxRadius=-1e10;

  forAll(oldPoints,pointI) {
    point oldPoint=oldPoints[pointI];
    point basePoint=cutPlane.nearestPoint(oldPoint);
    point axisPoint = ( basePoint & axisDir )*axisDir; 

    scalar radius=mag( axisPoint - basePoint );
    if(radius>maxRadius) {
        maxRadius=radius;
    }
    if(radius<minRadius) {
        minRadius=radius;
    }
    vector dir(oldPoint-basePoint);
#ifdef VERBOSE_OUTPUT
    Info << "Moving point " << oldPoint << " to ";
#endif
    newPoints[pointI] = basePoint + factor *radius * dir/mag(dir);
    oldPoints[pointI] = basePoint - factor *radius * dir/mag(dir);  //DPS rotating original plane

#ifdef VERBOSE_OUTPUT
    Info << newPoints[pointI] << endl; 
#endif
  }

  Info << "Radius to axis: min = " << minRadius << " max = " << maxRadius << endl;

  mesh.movePoints(oldPoints);
  mesh.movePoints(newPoints);
}

// Split the wedge-patch into two patches
 
void splitWedge(polyMesh &mesh,word wname,plane pl) {
  repatchPolyTopoChanger topo(mesh);

  const polyBoundaryMesh& patches = mesh.boundaryMesh();
  const polyPatch &wedge=patches[patches.findPatchID(wname)];
  
  const vectorField::subField 	& fcs=wedge.faceCentres ();

  faceSet facesPos(mesh,"set1_"+wname,fcs.size()/2,IOobject::NO_WRITE);
  faceSet facesNeg(mesh,"set2_"+wname,fcs.size()/2,IOobject::NO_WRITE);

  // assign faces to the two patches
  forAll(fcs,faceI) {
    vector center=fcs[faceI];

#ifdef VERBOSE_OUTPUT
    Info << " Face " << faceI-wedge.whichFace(0) << " " << center << " goes to ";
#endif

    scalar dir=pl.normal() & (center-pl.nearestPoint(center));
    if( dir >0 ) {
#ifdef VERBOSE_OUTPUT
      Info << facesPos.name();
#endif
      // This offset is dirty, there must be a better way
      facesPos.insert(faceI-wedge.whichFace(0));
    } else {
#ifdef VERBOSE_OUTPUT
      Info << facesNeg.name();
#endif
      facesNeg.insert(faceI-wedge.whichFace(0));
    }

#ifdef VERBOSE_OUTPUT
    Info << endl;
#endif
  }

  Info << " Copying patches " << endl;

  // the rest is stolen from createPatch (others might say inspired by)

  List<polyPatch*> newPatches(patches.size() + 2);
  
  forAll(patches, patchI)
    {
      const polyPatch& pp = patches[patchI];
      
      newPatches[patchI] =
	pp.clone
	(
	 patches,
	 patchI,
	 pp.size(),
	 pp.start()
	 ).ptr();
      
    }
  
  label patchPos = newPatches.size() - 2;
  label patchNeg = newPatches.size() - 1;
  
  Info << " Creating Patches" << endl;

  newPatches[patchPos] =
    polyPatch::New
    (
     "empty",
     wname+"_pos",
     0,
     mesh.nFaces(),
     patchPos,
     patches
     ).ptr();

   newPatches[patchNeg] =
     polyPatch::New
     (
      "empty",
      wname+"_neg",
      0,
      mesh.nFaces(),
      patchNeg,
      patches
      ).ptr();
  

  Info << " Creating Pos-patch " << endl;

  labelList faceLabelsPos(facesPos.toc());

  SortableList<label> patchFacesPos(faceLabelsPos);

  forAll(patchFacesPos, i)
    {
      label faceI = patchFacesPos[i];

      if (mesh.isInternalFace(faceI))
	{
	  FatalErrorIn("SplitWedge")
	    << "Face " << faceI << " specified in set " << facesPos.name()
	    << " is not an external face of the mesh." << endl
	    << "This application can only repatch existing boundary"
	    << " faces."
	    << exit(FatalError);
	}

      topo.changePatchID(patchFacesPos[i], patchPos);
    }

  Info << " Creating Neg-patch " << endl;

  labelList faceLabelsNeg(facesNeg.toc());

  SortableList<label> patchFacesNeg(faceLabelsNeg);

  forAll(patchFacesNeg, i)
    {
      label faceI = patchFacesNeg[i];

      if (mesh.isInternalFace(faceI))
	{
	  FatalErrorIn("SplitWedge")
	    << "Face " << faceI << " specified in set " << facesNeg.name()
	    << " is not an external face of the mesh." << endl
	    << "This application can only repatch existing boundary"
	    << " faces."
	    << exit(FatalError);
	}
      
      topo.changePatchID(patchFacesNeg[i], patchNeg);
    }

  Info << " Changing patches\n" << endl;

  // Actually add new list of patches
  topo.changePatches(newPatches);
  
  topo.repatch();

}

void changeTypes(polyMesh &mesh,word wedge,word axis) {
  repatchPolyTopoChanger topo(mesh);

  const polyBoundaryMesh& patches = mesh.boundaryMesh();

  List<polyPatch*> newPatches(patches.size());
  
  forAll(patches, patchI)
    {
      const polyPatch& pp = patches[patchI];

      if(pp.name()==axis) {
          Info << " Changing " << axis << " to empty" << endl;
          newPatches[patchI] =
              polyPatch::New(
                  "empty",
                  pp.name(),
                  pp.size(),
                  pp.start(),
                  patchI,
                  patches
              ).ptr();
      } else if(pp.name()==wedge+"_pos" || pp.name()==wedge+"_neg"){
          Info << " Changing " << pp.name() << " to wedge" << endl;
          newPatches[patchI] =
              polyPatch::New(
                  "wedge",
                  pp.name(),
                  pp.size(),
                  pp.start(),
                  patchI,
                  patches
              ).ptr();
      } else {
          newPatches[patchI] =
              pp.clone
              (
                  patches,
                  patchI,
                  pp.size(),
                  pp.start()
              ).ptr();
      }
    }

  Info << endl;

  topo.changePatches(newPatches);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
// Main program:

int main(int argc, char *argv[])
{
    argList::validArgs.append("axis face name");
    argList::validArgs.append("wedge face  name");
  
#   include "setRootCase.H"
#   include "createTime.H"

    Info<< "Create mesh for time = " << runTime.value() << nl << endl;

    polyMesh mesh
    (
        IOobject
        (
            polyMesh::defaultRegion,
            runTime.timeName(),
            runTime
        )
    );

    word axisName(args.args()[3]);
    word wedgeName(args.args()[4]);

    pointField& points = const_cast<pointField&>(mesh.points());

    plane approx=getPlane(points);

    Info << "Plane of the grid: " << approx << "\n" << endl;

    const polyBoundaryMesh& patches = mesh.boundaryMesh();

    if (patches.findPatchID(axisName) == -1)
    {
        FatalErrorIn(args.executable())
            << "Patch " << axisName << " does not exists in mesh" << endl
            << "Patches are " << patches.names()
            << exit(FatalError);
    }

    if (patches.findPatchID(wedgeName) == -1)
    {
        FatalErrorIn(args.executable())
            << "Patch " << wedgeName << " does not exists in mesh" << endl
            << "Patches are " << patches.names()
            << exit(FatalError);
    }

    if( axisName == wedgeName) {
        FatalErrorIn(args.executable())
	  << "Patch " << axisName << " can't be used as axis and wedge patch" << endl
	  << exit(FatalError);	
    }

    const polyPatch &axisPatch=patches[patches.findPatchID(axisName)];

    linie theAxis=getAxis(axisPatch,approx);

    Info << "The rotation-axis: " << theAxis << "\n" << endl;

    Info << "Creating wedge with an opening angle of " << angle << " degrees\n" << endl;

    changeCoordinates(mesh,approx,theAxis);

    Info << "Splitting patch " << wedgeName << "\n" << endl;

    splitWedge(mesh,wedgeName,approx);

    Info << "Changing patch types " << endl;

    changeTypes(mesh,wedgeName,axisName);

    runTime++;
    
    Info << "Writing mesh to time " << runTime.value() << endl;
    mesh.write();

    Info << "End\n" << endl;

    Info << "Now use collapseEdges to clean grid\n";

    return 0;
}


// ************************************************************************* //
